@@ -22,7 +22,24 @@ from pycyphal.transport.can.media import Media, FilterConfiguration, Envelope, F
 _logger = logging.getLogger(__name__)
 
 
+@dataclasses.dataclass(frozen=True)
+class PythonCANBusOptions:
+    hardware_loopback: bool = False
+    """
+    Hardware loopback support.
+        If True, loopback is handled by the supported hardware.
+        If False, loopback is emulated with software.
+    """
+    hardware_timestamp: bool = False
+    """
+    Hardware timestamp support.
+        If True, timestamp returned by the hardware is used.
+        If False, approximate timestamp is captured by software.
+    """
+
+
 class PythonCANMedia(Media):
+    # pylint: disable=line-too-long
     """
     Media interface adapter for `Python-CAN <https://python-can.readthedocs.io/>`_.
     It is designed to be usable with all host platforms supported by Python-CAN (GNU/Linux, Windows, macOS).
@@ -31,22 +48,20 @@ class PythonCANMedia(Media):
 
     This media interface supports both Classic CAN and CAN FD. The selection logic is documented below.
 
-    This implementation emulates loopback instead of requesting it from the underlying driver due to the limitations
-    of Python-CAN. Same goes for timestamping.
-    If accurate timestamping is desired, consider using the SocketCAN media driver instead.
+    Python-CAN supports hardware loopback and timestamping only for some of the interfaces. This has to be manually
+    specified in PythonCANBusOptions for supported hardware. Both are disabled by default, but can be enabled if it
+    is verified that hardware in question supports either or both options.
+    For best compatibility, consider using the non-python-can SocketCAN media driver instead.
 
     Here is a basic usage example based on the Yakut CLI tool.
     Suppose that there are two interconnected CAN bus adapters connected to the host computer:
     one SLCAN-based, the other is PCAN USB.
-    Launch Yakut to listen for messages using the SLCAN adapter::
+    Launch Yakut to listen for messages using the SLCAN adapter (only one at a time)::
 
-        export YAKUT_TRANSPORT='CAN(can.media.pythoncan.PythonCANMedia("slcan:/dev/ttyACM0", 1_000_000), None)'
-        yakut sub 33.uavcan.si.unit.voltage.Scalar.1.0
-
-    While the subscriber is running, publish a message to this subject::
-
-        export YAKUT_TRANSPORT='CAN(can.media.pythoncan.PythonCANMedia("pcan:PCAN_USBBUS1", 1_000_000), 42)'
-        yakut pub 33.uavcan/si/unit/voltage/Scalar_1_0 '{volt: 12}'
+        export UAVCAN__CAN__IFACE="slcan:/dev/serial/by-id/usb-Zubax_Robotics_Zubax_Babel_1B003D00145130365030332000000000-if00"
+        export UAVCAN__CAN__BITRATE='1000000 1000000'
+        export UAVCAN__CAN__MTU=8
+        yakut sub 33:uavcan.si.unit.voltage.scalar
     """
 
     _MAXIMAL_TIMEOUT_SEC = 0.1
@@ -99,6 +114,11 @@ class PythonCANMedia(Media):
               https://python-can.readthedocs.io/en/stable/interfaces/seeedstudio.html.
               Example: ``seeedstudio:/dev/ttyUSB0`` (Linux) or ``seeedstudio:COM3`` (Windows)
 
+            - Interface ``gs_usb`` is implemented by :class:`can.interfaces.gs_usb.GsUsbBus`.
+              Channel name is an integer, refering to the device index in a system.
+              Example: ``gs_usb:0``
+              Note: this interface currently requires unreleased `python-can` version from git.
+
         :param bitrate: Bit rate value in bauds; either a single integer or a tuple:
 
             - A single integer selects Classic CAN.
@@ -180,7 +200,9 @@ class PythonCANMedia(Media):
                 interface_name=self._conn_name[0], channel_name=self._conn_name[1], bitrate=bitrate[0]
             )
         try:
-            self._bus = _CONSTRUCTORS[self._conn_name[0]](params)
+            bus_options, bus = _CONSTRUCTORS[self._conn_name[0]](params)
+            self._bus_options: PythonCANBusOptions = bus_options
+            self._bus: can.ThreadSafeBus = bus
         except can.CanError as ex:
             raise InvalidMediaConfigurationError(f"Could not initialize PythonCAN: {ex}") from ex
         super().__init__()
@@ -246,9 +268,10 @@ class PythonCANMedia(Media):
                 is_fd=self._is_fd,
             )
             try:
+                desired_timeout = monotonic_deadline - loop.time()
                 await loop.run_in_executor(
                     self._background_executor,
-                    functools.partial(self._bus.send, message, timeout=monotonic_deadline - loop.time()),
+                    functools.partial(self._bus.send, message, timeout=max(desired_timeout, 0)),
                 )
             except (asyncio.TimeoutError, can.CanError):  # CanError is also used to report timeouts (weird).
                 break
@@ -256,7 +279,8 @@ class PythonCANMedia(Media):
                 num_sent += 1
                 if f.loopback:
                     loopback.append((Timestamp.now(), f))
-        if loopback:
+        # Fake received frames if hardware does not support loopback
+        if loopback and not self._bus_options.hardware_loopback:
             loop.call_soon(self._invoke_rx_handler, loopback)
         return num_sent
 
@@ -315,8 +339,12 @@ class PythonCANMedia(Media):
             msg = self._bus.recv(0.0 if batch else self._MAXIMAL_TIMEOUT_SEC)
             if msg is None:
                 break
-            timestamp = Timestamp.now()  # TODO: use accurate timestamping
-            loopback = False  # TODO: no possibility to get real loopback yet
+
+            mono_ns = msg.timestamp * 1e9 if self._bus_options.hardware_timestamp else time.monotonic_ns()
+            timestamp = Timestamp(system_ns=time.time_ns(), monotonic_ns=mono_ns)
+
+            loopback = self._bus_options.hardware_loopback and (not msg.is_rx)
+
             frame = self._parse_native_frame(msg)
             if frame is not None:
                 batch.append((timestamp, Envelope(frame, loopback)))
@@ -350,37 +378,52 @@ class _FDInterfaceParameters(_InterfaceParameters):
 
 def _construct_socketcan(parameters: _InterfaceParameters) -> can.ThreadSafeBus:
     if isinstance(parameters, _ClassicInterfaceParameters):
-        return can.ThreadSafeBus(interface=parameters.interface_name, channel=parameters.channel_name, fd=False)
+        return (
+            PythonCANBusOptions(),
+            can.ThreadSafeBus(interface=parameters.interface_name, channel=parameters.channel_name, fd=False),
+        )
     if isinstance(parameters, _FDInterfaceParameters):
-        return can.ThreadSafeBus(interface=parameters.interface_name, channel=parameters.channel_name, fd=True)
+        return (
+            PythonCANBusOptions(),
+            can.ThreadSafeBus(interface=parameters.interface_name, channel=parameters.channel_name, fd=True),
+        )
     assert False, "Internal error"
 
 
 def _construct_kvaser(parameters: _InterfaceParameters) -> can.ThreadSafeBus:
     if isinstance(parameters, _ClassicInterfaceParameters):
-        return can.ThreadSafeBus(
-            interface=parameters.interface_name,
-            channel=parameters.channel_name,
-            bitrate=parameters.bitrate,
-            fd=False,
+        return (
+            PythonCANBusOptions(),
+            can.ThreadSafeBus(
+                interface=parameters.interface_name,
+                channel=parameters.channel_name,
+                bitrate=parameters.bitrate,
+                fd=False,
+            ),
         )
     if isinstance(parameters, _FDInterfaceParameters):
-        return can.ThreadSafeBus(
-            interface=parameters.interface_name,
-            channel=parameters.channel_name,
-            bitrate=parameters.bitrate[0],
-            fd=True,
-            data_bitrate=parameters.bitrate[1],
+        return (
+            PythonCANBusOptions(),
+            can.ThreadSafeBus(
+                interface=parameters.interface_name,
+                channel=parameters.channel_name,
+                bitrate=parameters.bitrate[0],
+                fd=True,
+                data_bitrate=parameters.bitrate[1],
+            ),
         )
     assert False, "Internal error"
 
 
 def _construct_slcan(parameters: _InterfaceParameters) -> can.ThreadSafeBus:
     if isinstance(parameters, _ClassicInterfaceParameters):
-        return can.ThreadSafeBus(
-            interface=parameters.interface_name,
-            channel=parameters.channel_name,
-            bitrate=parameters.bitrate,
+        return (
+            PythonCANBusOptions(),
+            can.ThreadSafeBus(
+                interface=parameters.interface_name,
+                channel=parameters.channel_name,
+                bitrate=parameters.bitrate,
+            ),
         )
     if isinstance(parameters, _FDInterfaceParameters):
         raise InvalidMediaConfigurationError(f"Interface does not support CAN FD: {parameters.interface_name}")
@@ -389,10 +432,13 @@ def _construct_slcan(parameters: _InterfaceParameters) -> can.ThreadSafeBus:
 
 def _construct_pcan(parameters: _InterfaceParameters) -> can.ThreadSafeBus:
     if isinstance(parameters, _ClassicInterfaceParameters):
-        return can.ThreadSafeBus(
-            interface=parameters.interface_name,
-            channel=parameters.channel_name,
-            bitrate=parameters.bitrate,
+        return (
+            PythonCANBusOptions(),
+            can.ThreadSafeBus(
+                interface=parameters.interface_name,
+                channel=parameters.channel_name,
+                bitrate=parameters.bitrate,
+            ),
         )
     if isinstance(parameters, _FDInterfaceParameters):
         # These magic numbers come from the settings of PCAN adapter.
@@ -407,19 +453,22 @@ def _construct_pcan(parameters: _InterfaceParameters) -> can.ThreadSafeBus:
         data_br = int(f_clock / parameters.bitrate[1] / (data_tseg1 + data_tseg2 + data_sjw))
         # TODO: validate the result and see if it is within an acceptable range
 
-        return can.ThreadSafeBus(
-            interface=parameters.interface_name,
-            channel=parameters.channel_name,
-            f_clock=f_clock,
-            nom_brp=nom_br,
-            data_brp=data_br,
-            nom_tseg1=nom_tseg1,
-            nom_tseg2=nom_tseg2,
-            nom_sjw=nom_sjw,
-            data_tseg1=data_tseg1,
-            data_tseg2=data_tseg2,
-            data_sjw=data_sjw,
-            fd=True,
+        return (
+            PythonCANBusOptions(),
+            can.ThreadSafeBus(
+                interface=parameters.interface_name,
+                channel=parameters.channel_name,
+                f_clock=f_clock,
+                nom_brp=nom_br,
+                data_brp=data_br,
+                nom_tseg1=nom_tseg1,
+                nom_tseg2=nom_tseg2,
+                nom_sjw=nom_sjw,
+                data_tseg1=data_tseg1,
+                data_tseg2=data_tseg2,
+                data_sjw=data_sjw,
+                fd=True,
+            ),
         )
 
     assert False, "Internal error"
@@ -427,18 +476,24 @@ def _construct_pcan(parameters: _InterfaceParameters) -> can.ThreadSafeBus:
 
 def _construct_virtual(parameters: _InterfaceParameters) -> can.ThreadSafeBus:
     if isinstance(parameters, _ClassicInterfaceParameters):
-        return can.ThreadSafeBus(interface=parameters.interface_name, bitrate=parameters.bitrate)
+        return (
+            PythonCANBusOptions(),
+            can.ThreadSafeBus(interface=parameters.interface_name, bitrate=parameters.bitrate),
+        )
     if isinstance(parameters, _FDInterfaceParameters):
-        return can.ThreadSafeBus(interface=parameters.interface_name)
+        return (PythonCANBusOptions(), can.ThreadSafeBus(interface=parameters.interface_name))
     assert False, "Internal error"
 
 
 def _construct_usb2can(parameters: _InterfaceParameters) -> can.ThreadSafeBus:
     if isinstance(parameters, _ClassicInterfaceParameters):
-        return can.ThreadSafeBus(
-            interface=parameters.interface_name,
-            channel=parameters.channel_name,
-            bitrate=parameters.bitrate,
+        return (
+            PythonCANBusOptions(),
+            can.ThreadSafeBus(
+                interface=parameters.interface_name,
+                channel=parameters.channel_name,
+                bitrate=parameters.bitrate,
+            ),
         )
     if isinstance(parameters, _FDInterfaceParameters):
         raise InvalidMediaConfigurationError(f"Interface does not support CAN FD: {parameters.interface_name}")
@@ -447,8 +502,11 @@ def _construct_usb2can(parameters: _InterfaceParameters) -> can.ThreadSafeBus:
 
 def _construct_canalystii(parameters: _InterfaceParameters) -> can.ThreadSafeBus:
     if isinstance(parameters, _ClassicInterfaceParameters):
-        return can.ThreadSafeBus(
-            interface=parameters.interface_name, channel=parameters.channel_name, bitrate=parameters.bitrate
+        return (
+            PythonCANBusOptions(),
+            can.ThreadSafeBus(
+                interface=parameters.interface_name, channel=parameters.channel_name, bitrate=parameters.bitrate
+            ),
         )
     if isinstance(parameters, _FDInterfaceParameters):
         raise InvalidMediaConfigurationError(f"Interface does not support CAN FD: {parameters.interface_name}")
@@ -457,22 +515,50 @@ def _construct_canalystii(parameters: _InterfaceParameters) -> can.ThreadSafeBus
 
 def _construct_seeedstudio(parameters: _InterfaceParameters) -> can.ThreadSafeBus:
     if isinstance(parameters, _ClassicInterfaceParameters):
-        return can.ThreadSafeBus(
-            interface=parameters.interface_name,
-            channel=parameters.channel_name,
-            bitrate=parameters.bitrate,
+        return (
+            PythonCANBusOptions(),
+            can.ThreadSafeBus(
+                interface=parameters.interface_name,
+                channel=parameters.channel_name,
+                bitrate=parameters.bitrate,
+            ),
         )
     if isinstance(parameters, _FDInterfaceParameters):
         raise InvalidMediaConfigurationError(f"Interface does not support CAN FD: {parameters.interface_name}")
     assert False, "Internal error"
 
 
+def _construct_gs_usb(parameters: _InterfaceParameters) -> can.ThreadSafeBus:
+    if isinstance(parameters, _ClassicInterfaceParameters):
+        try:
+            index = int(parameters.channel_name)
+        except ValueError:
+            raise InvalidMediaConfigurationError("Channel name must be an integer interface index") from None
+
+        try:
+            bus = can.ThreadSafeBus(
+                interface=parameters.interface_name,
+                channel=parameters.channel_name,
+                index=index,
+                bitrate=parameters.bitrate,
+            )
+        except TypeError as e:
+            raise InvalidMediaConfigurationError(
+                f"Interface error: {e}.\nNote: gs_usb currently requires unreleased python-can version from git."
+            ) from e
+
+        return (PythonCANBusOptions(hardware_loopback=True, hardware_timestamp=True), bus)
+    if isinstance(parameters, _FDInterfaceParameters):
+        raise InvalidMediaConfigurationError(f"Interface does not support CAN FD: {parameters.interface_name}")
+    assert False, "Internal error"
+
+
 def _construct_any(parameters: _InterfaceParameters) -> can.ThreadSafeBus:
     raise InvalidMediaConfigurationError(f"Interface not supported yet: {parameters.interface_name}")
 
 
 _CONSTRUCTORS: typing.DefaultDict[
-    str, typing.Callable[[_InterfaceParameters], can.ThreadSafeBus]
+    str, typing.Callable[[_InterfaceParameters], typing.Tuple[PythonCANBusOptions, can.ThreadSafeBus]]
 ] = collections.defaultdict(
     lambda: _construct_any,
     {
@@ -484,5 +570,6 @@ _CONSTRUCTORS: typing.DefaultDict[
         "usb2can": _construct_usb2can,
         "canalystii": _construct_canalystii,
         "seeedstudio": _construct_seeedstudio,
+        "gs_usb": _construct_gs_usb,
     },
 )
